/*
 *
 * Pyldin-601 emulator version 2.1 for BeOS / Linux / Win9x/NT
 * Copyright (c) Sasha Chukov & Yura Kuznetsov, 2000-2001
 *
 */

#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <conio.h>

#include "dpmi.h"
#include "floppy.h"

#define SSIZE	512
#define NSECT	9
#define NHEAD	2

SELECTOR sel = NULL;
UWORD seg = NULL;

UBYTE *buffer;

char *diskImage[] = {
	NULL, NULL, NULL, NULL
};

int dSizes[] = {
	0, 0, 0, 0
};

int flopAssoc[] = {
	0, 1, 2, 3
};

int floppyOn()
{
	if (dpmi_alloc_real_memory(&sel, &seg, 1024>>4) == 0) {
		buffer = (UBYTE *)(seg << 4);
		return 0;
	} else return -1;
}

int floppyOff()
{
	if (sel)
	   if (dpmi_dealloc_real_memory(&sel) == 0) return 0;
	   else { sel = NULL; return -1; }
	return 0;
}

int floppyOp(int Op, int Drive, int Track, int Head, int Sector, UBYTE *mema)
{
  DPMI_REGS regs;
  int retc = 0;

  Drive = flopAssoc[Drive];

//  Op &= 0x03;

  if (!sel) return 0xc0;

  switch( Op ){
	case 0: regs.b.ah = 0x0; break;
	case 1: regs.b.ah = 0x2; break;
	case 2: regs.b.ah = 0x3; memcpy(buffer, mema, SSIZE); break;
//	case 4: return 0;//;regs.b.ah = 0x3; memset(buffer, 0xe6, SSIZE); break;
	default: return 0; break;
  }

  regs.b.al = 0x01;
  regs.b.ch = Track;
  regs.b.cl = Sector;
  regs.b.dh = Head;
  regs.b.dl = Drive;

  regs.h.bx = 0;
  regs.h.es = seg;
  regs.h.flags = 0x72;
  regs.h.ss = 0;
  regs.h.sp = 0;

	_disable();
	outp(0x43, 0x36);
	outp(0x40, 0x00);
	outp(0x40, 0x00);
	_enable();

  dpmi_simulate_real_interrupt(0x13,&regs);

	_disable();
	outp(0x43, 0x36);
	outp(0x40, 0x30);
	outp(0x40, 0x56);
	_enable();

  if (Op == 1) memcpy(mema, buffer, SSIZE);

  switch(regs.b.ah){
	case 0x00: retc = 0x00; break;
	case 0x02: retc = 0x10; break;
	case 0x03: retc = 0x01; break;
	case 0x08: retc = 0x02; break;
	case 0x04:
	case 0x0c:
	case 0x40: retc = 0x20; break;
	case 0x10: retc = 0x04; break;
	case 0x20:
	case 0xaa: retc = 0x80; break;
	case 0x80: retc = 0x40; break;
	case 0x0a:
	case 0x0b: retc = 0x08; break;
	default: retc = 0xc0; break;
  }

  return retc;
}

int loadDiskImage(char *name, int disk)
{
	if (strcmp("A:", name)==0) { flopAssoc[disk] = 0; return 0; }
	else if (strcmp("a:", name)==0) { flopAssoc[disk] = 0; return 0; }
	else if (strcmp("B:", name)==0) { flopAssoc[disk] = 1; return 0; }
	else if (strcmp("b:", name)==0) { flopAssoc[disk] = 1; return 0; }

	FILE *inf = fopen(name, "rb");
	if (!inf) return -1;
	if (!diskImage[disk]) diskImage[disk] = new char[737280*2];
	dSizes[disk] = fread(diskImage[disk], 1, 737280*2, inf);
	fclose(inf);
	return 0;
}

int unloadDiskImage(char *name, int disk)
{
	if (!diskImage[disk]) return -1;
	FILE *outf = fopen(name, "wb");
	if (!outf) return -1;
	fwrite(diskImage[disk], 1, dSizes[disk], outf);
	if (diskImage[disk]) delete diskImage[disk];
	fclose(outf);
	return 0;
}

int readSector(int Disk, int Track, int Sector, int Head, unsigned char *dst)
{
	if (Track > 79 || Sector > 18) return 0x40;
	if (!diskImage[Disk])
		return floppyOp(1, Disk, Track, Head, Sector, dst);

	int offs;
	if (dSizes[Disk] > 737280) offs = (Sector-1)*SSIZE+Head*18*SSIZE+Track*18*NHEAD*SSIZE;
	else offs = (Sector-1)*SSIZE+Head*NSECT*SSIZE+Track*NSECT*NHEAD*SSIZE;

	memcpy(dst, diskImage[Disk] + offs, SSIZE);
	return 0;
}

int writeSector(int Disk, int Track, int Sector, int Head, unsigned char *src)
{
	if (Track > 79 || Sector > 18) return 0x40;
	if (!diskImage[Disk])
		return floppyOp(2, Disk, Track, Head, Sector, src);

	int offs;
	if (dSizes[Disk] > 737280) offs = (Sector-1)*SSIZE+Head*18*SSIZE+Track*18*NHEAD*SSIZE;
	else offs = (Sector-1)*SSIZE+Head*NSECT*SSIZE+Track*NSECT*NHEAD*SSIZE;

	memcpy(diskImage[Disk] + offs, src, SSIZE);
	return 0;
}

int formaTrack(int Disk, int Track, int Head)
{
	if (Track > 79) return 0x40;
	if (!diskImage[Disk])
		return floppyOp(4, Disk, Track, Head, NULL, NULL);

	int offs;
	if (dSizes[Disk] > 737280) offs = Head*18*SSIZE+Track*18*NHEAD*SSIZE;
	else offs = Head*NSECT*SSIZE+Track*NSECT*NHEAD*SSIZE;

	if (dSizes[Disk] > 737280) memset(diskImage[Disk] + offs, 0xf6, 18*SSIZE);
	else memset(diskImage[Disk] + offs, 0xe5, NSECT*SSIZE);
	return 0;
}

int init765()
{
  if (!diskImage[FLOPPY_A])
    floppyOp(0, FLOPPY_A, NULL, NULL, NULL, NULL);
  if (!diskImage[FLOPPY_B])
    floppyOp(0, FLOPPY_B, NULL, NULL, NULL, NULL);
  return 0;
}
